@@ -16,7 +16,8 @@
 #include "ftcommon.h"
 #include "common.h"
 #include "output.h"
-#include <math.h>
+#include "mlgetopt.h"
+#include <stdlib.h>
 
   /* the following header shouldn't be used in normal programs */
 #include FT_INTERNAL_DEBUG_H
@@ -28,6 +29,7 @@
 
 #include FT_STROKER_H
 #include FT_SYNTHESIS_H
+#include FT_LCD_FILTER_H
 #include FT_CFF_DRIVER_H
 #include FT_TRUETYPE_DRIVER_H
 #include FT_MULTIPLE_MASTERS_H
@@ -89,6 +91,10 @@
 
 #define BUFSIZE  256
 
+#define DO_BITMAP      1
+#define DO_OUTLINE     2
+#define DO_DOTS        4
+#define DO_DOTNUMBERS  8
 
   typedef struct  GridStatusRec_
   {
@@ -100,14 +106,13 @@
     int          Num;  /* glyph index */
     int          font_index;
 
-    double       scale;
-    double       x_origin;
-    double       y_origin;
-    double       margin;
+    FT_F26Dot6   scale;
+    int          x_origin;
+    int          y_origin;
 
-    double       scale_0;
-    double       x_origin_0;
-    double       y_origin_0;
+    FT_F26Dot6   scale_0;
+    int          x_origin_0;
+    int          y_origin_0;
 
     int          disp_width;
     int          disp_height;
@@ -121,27 +126,30 @@
     grColor      segment_color;
     grColor      blue_color;
 
+    int          work;
     int          do_horz_hints;
     int          do_vert_hints;
     int          do_blue_hints;
-    int          do_outline;
-    int          do_dots;
-    int          do_dotnumbers;
     int          do_segment;
 
+    FT_LcdFilter lcd_filter;
     double       gamma;
     const char*  header;
     char         header_buffer[BUFSIZE];
 
-    FT_Stroker  stroker;
+    FT_Stroker   stroker;
 
     unsigned int cff_hinting_engine;
-    unsigned int tt_interpreter_version;
+    unsigned int tt_interpreter_versions[3];
+    int          num_tt_interpreter_versions;
+    int          tt_interpreter_version_idx;
     FT_Bool      warping;
 
     FT_MM_Var*   mm;
     char*        axis_name[MAX_MM_AXES];
     FT_Fixed     design_pos[MAX_MM_AXES];
+    FT_Fixed     requested_pos[MAX_MM_AXES];
+    FT_UInt      requested_cnt;
     FT_UInt      current_axis;
     FT_UInt      used_num_axis;
 
@@ -156,22 +164,20 @@
     st->width         = DIM_X;
     st->height        = DIM_Y;
 
-    st->scale         = 1.0;
+    st->scale         = 64;
     st->x_origin      = 0;
     st->y_origin      = 0;
-    st->margin        = 0.05;
 
+    st->work          = DO_BITMAP | DO_OUTLINE | DO_DOTS;
     st->do_horz_hints = 1;
     st->do_vert_hints = 1;
     st->do_blue_hints = 1;
-    st->do_dots       = 1;
-    st->do_dotnumbers = 0;
-    st->do_outline    = 1;
     st->do_segment    = 0;
 
     st->Num           = 0;
+    st->lcd_filter    = FT_LCD_FILTER_DEFAULT;
     st->gamma         = GAMMA;
-    st->header        = "";
+    st->header        = NULL;
 
     st->mm            = NULL;
     st->current_axis  = 0;
@@ -200,8 +206,9 @@
   grid_status_rescale_initial( GridStatus      st,
                                FTDemo_Handle*  handle )
   {
-    FT_Size   size;
-    FT_Error  err = FTDemo_Get_Size( handle, &size );
+    FT_Size     size;
+    FT_Error    err    = FTDemo_Get_Size( handle, &size );
+    FT_F26Dot6  margin = 4;
 
 
     if ( !err )
@@ -213,7 +220,7 @@
       int  xmax = FT_MulFix( face->bbox.xMax, size->metrics.x_scale );
       int  ymax = FT_MulFix( face->bbox.yMax, size->metrics.y_scale );
 
-      double  x_scale, y_scale;
+      FT_F26Dot6  x_scale, y_scale;
 
 
       xmin &= ~63;
@@ -222,30 +229,33 @@
       ymax  = ( ymax + 63 ) & ~63;
 
       if ( xmax - xmin )
-        x_scale = st->disp_width  * ( 1.0 - 2 * st->margin ) / ( xmax - xmin );
+        x_scale = st->disp_width  * ( 64 - 2 * margin ) / ( xmax - xmin );
       else
-        x_scale = 1.0;
+        x_scale = 64;
 
       if ( ymax - ymin )
-        y_scale = st->disp_height * ( 1.0 - 2 * st->margin ) / ( ymax - ymin );
+        y_scale = st->disp_height * ( 64 - 2 * margin ) / ( ymax - ymin );
       else
-        y_scale = 1.0;
+        y_scale = 64;
 
       if ( x_scale <= y_scale )
         st->scale = x_scale;
       else
         st->scale = y_scale;
 
-      st->x_origin = st->disp_width  * st->margin         - xmin * st->scale;
-      st->y_origin = st->disp_height * ( 1 - st->margin ) + ymin * st->scale;
+      st->x_origin = st->disp_width  * margin          - xmin * st->scale;
+      st->y_origin = st->disp_height * ( 64 - margin ) + ymin * st->scale;
     }
     else
     {
-      st->scale    = 1.;
-      st->x_origin = st->disp_width  * st->margin;
-      st->y_origin = st->disp_height * st->margin;
+      st->scale    = 64;
+      st->x_origin = st->disp_width  * margin;
+      st->y_origin = st->disp_height * ( 64 - margin );
     }
 
+    st->x_origin >>= 6;
+    st->y_origin >>= 6;
+
     st->scale_0    = st->scale;
     st->x_origin_0 = st->x_origin;
     st->y_origin_0 = st->y_origin;
@@ -255,31 +265,31 @@
   static void
   grid_status_draw_grid( GridStatus  st )
   {
-    int     x_org   = (int)st->x_origin;
-    int     y_org   = (int)st->y_origin;
-    double  xy_incr = 64.0 * st->scale;
+    int         x_org   = st->x_origin;
+    int         y_org   = st->y_origin;
+    FT_F26Dot6  xy_incr = st->scale;
 
 
-    if ( xy_incr >= 2. )
+    if ( xy_incr >= 2 )
     {
-      double  x2 = x_org;
-      double  y2 = y_org;
+      int  x2 = x_org;
+      int  y2 = y_org;
 
 
       for ( ; x2 < st->disp_width; x2 += xy_incr )
-        grFillVLine( st->disp_bitmap, (int)x2, 0,
+        grFillVLine( st->disp_bitmap, x2, 0,
                      st->disp_height, st->grid_color );
 
-      for ( x2 = x_org - xy_incr; (int)x2 >= 0; x2 -= xy_incr )
-        grFillVLine( st->disp_bitmap, (int)x2, 0,
+      for ( x2 = x_org - xy_incr; x2 >= 0; x2 -= xy_incr )
+        grFillVLine( st->disp_bitmap, x2, 0,
                      st->disp_height, st->grid_color );
 
       for ( ; y2 < st->disp_height; y2 += xy_incr )
-        grFillHLine( st->disp_bitmap, 0, (int)y2,
+        grFillHLine( st->disp_bitmap, 0, y2,
                      st->disp_width, st->grid_color );
 
-      for ( y2 = y_org - xy_incr; (int)y2 >= 0; y2 -= xy_incr )
-        grFillHLine( st->disp_bitmap, 0, (int)y2,
+      for ( y2 = y_org - xy_incr; y2 >= 0; y2 -= xy_incr )
+        grFillHLine( st->disp_bitmap, 0, y2,
                      st->disp_width, st->grid_color );
     }
 
@@ -297,8 +307,8 @@
                           AF_GlyphHints  hints )
   {
     FT_Int  dimension;
-    int     x_org = (int)st->x_origin;
-    int     y_org = (int)st->y_origin;
+    int     x_org = st->x_origin;
+    int     y_org = st->y_origin;
 
 
     for ( dimension = 1; dimension >= 0; dimension-- )
@@ -323,17 +333,17 @@
 
         if ( dimension == 0 ) /* AF_DIMENSION_HORZ is 0 */
         {
-          pos = x_org + (int)( offset * st->scale );
+          pos = x_org + ( ( offset * st->scale ) >> 6 );
           grFillVLine( st->disp_bitmap, pos, 0,
                        st->disp_height, st->segment_color );
         }
         else
         {
-          pos = y_org - (int)( offset * st->scale );
+          pos = y_org - ( ( offset * st->scale ) >> 6 );
 
           if ( is_blue )
           {
-            int  blue_pos = y_org - (int)( blue_offset * st->scale );
+            int  blue_pos = y_org - ( ( blue_offset * st->scale ) >> 6 );
 
 
             if ( blue_pos == pos )
@@ -474,7 +484,8 @@
   {
     char*       tag;
     FT_Vector*  vec;
-    FT_F26Dot6  disp = (FT_F26Dot6)( radius * 0.6781 );
+    FT_F26Dot6  disp = (FT_F26Dot6)( radius * 0.5523 );
+    /* so that Bézier curve touches circle at 0, 45, and 90 degrees */
 
 
     FT_Outline_New( handle->library, 12, 1, outline );
@@ -512,7 +523,8 @@
 
 
     ft_outline_new_circle( &outline, radius, handle );
-    FT_Outline_Translate( &outline, center_x & 63, center_y & 63 );
+    /* subpixel adjustment considering downward direction of y-axis */
+    FT_Outline_Translate( &outline, center_x & 63, -( center_y & 63 ) );
 
     ft_outline_draw( &outline, 1., ( center_x >> 6 ), ( center_y >> 6 ),
                      handle, display, color );
@@ -521,6 +533,108 @@
   }
 
 
+  static void
+  bitmap_scale( grBitmap*   bit,
+                FT_F26Dot6  scale )
+  {
+    unsigned char*  s = bit->buffer;
+    unsigned char*  t;
+    unsigned char*  line;
+    int             pitch;
+    int             width;
+    int             i, j, k;
+
+    pitch = bit->pitch > 0 ?  bit->pitch
+                           : -bit->pitch;
+    width = bit->width;
+
+    t = (unsigned char*)malloc( (size_t)( pitch * bit->rows *
+                                          scale * scale ) );
+    if ( !t )
+      return;
+
+    line = t;
+
+    switch( bit->mode )
+    {
+      case gr_pixel_mode_mono:
+        for ( i = 0; i < bit->rows; i++ )
+        {
+          for ( j = 0; j < pitch * scale * 8; j++ )
+            if ( s[i * pitch + j / scale / 8] & ( 0x80 >> ( j / scale & 7 ) ) )
+              line[j / 8] |= 0x80 >> ( j & 7 );
+            else
+              line[j / 8] &= ~( 0x80 >> ( j & 7 ) );
+
+          for ( k = 1; k < scale; k++, line += pitch * scale )
+            memcpy( line + pitch * scale, line, (size_t)( pitch * scale ) );
+          line += pitch * scale;
+        }
+        break;
+
+      case gr_pixel_mode_gray:
+        for ( i = 0; i < bit->rows; i++ )
+        {
+          for ( j = 0; j < pitch; j++ )
+            memset( line + j * scale, s[i * pitch + j], (size_t)scale );
+
+          for ( k = 1; k < scale; k++, line += pitch * scale )
+            memcpy( line + pitch * scale, line, (size_t)( pitch * scale ) );
+          line += pitch * scale;
+        }
+        break;
+
+      case gr_pixel_mode_lcd:
+      case gr_pixel_mode_lcd2:
+        for ( i = 0; i < bit->rows; i++ )
+        {
+          for ( j = 0; j < width; j += 3 )
+            for ( k = 0; k < scale; k++ )
+            {
+              line[j * scale + 3 * k    ] = s[i * pitch + j    ];
+              line[j * scale + 3 * k + 1] = s[i * pitch + j + 1];
+              line[j * scale + 3 * k + 2] = s[i * pitch + j + 2];
+            }
+
+          for ( k = 1; k < scale; k++, line += pitch * scale )
+            memcpy( line + pitch * scale, line, (size_t)( pitch * scale ) );
+          line += pitch * scale;
+        }
+        break;
+
+      case gr_pixel_mode_lcdv:
+      case gr_pixel_mode_lcdv2:
+        for ( i = 0; i < bit->rows; i += 3 )
+        {
+          for ( j = 0; j < pitch; j++ )
+          {
+            memset( line + j * scale,
+                    s[i * pitch +             j], (size_t)scale );
+            memset( line + j * scale +     pitch * scale,
+                    s[i * pitch +     pitch + j], (size_t)scale );
+            memset( line + j * scale + 2 * pitch * scale,
+                    s[i * pitch + 2 * pitch + j], (size_t)scale );
+          }
+
+          for ( k = 1; k < scale; k++, line += 3 * pitch * scale )
+            memcpy( line + 3 * pitch * scale,
+                    line,
+                    (size_t)( 3 * pitch * scale ) );
+          line += 3 * pitch * scale;
+        }
+        break;
+
+      default:
+        return;
+    }
+
+    bit->buffer = t;
+    bit->rows  *= scale;
+    bit->width *= scale;
+    bit->pitch *= scale;
+  }
+
+
   static void
   grid_status_draw_outline( GridStatus       st,
                             FTDemo_Handle*   handle,
@@ -528,9 +642,9 @@
   {
     FT_Size       size;
     FT_GlyphSlot  slot;
-    double        scale = 64.0 * st->scale;
-    int           ox    = (int)st->x_origin;
-    int           oy    = (int)st->y_origin;
+    FT_F26Dot6    scale = st->scale;
+    int           ox    = st->x_origin;
+    int           oy    = st->y_origin;
 
 
     if ( st->stroker == NULL )
@@ -538,7 +652,7 @@
       FT_Stroker_New( handle->library, &st->stroker );
 
       FT_Stroker_Set( st->stroker, 32, FT_STROKER_LINECAP_BUTT,
-                      FT_STROKER_LINEJOIN_ROUND, 0x20000 );
+                      FT_STROKER_LINEJOIN_BEVEL, 0x20000 );
     }
 
     FTDemo_Get_Size( handle, &size );
@@ -570,9 +684,11 @@
 
     /* show advance width */
     grFillVLine( st->disp_bitmap,
-                 (int)( st->x_origin +
-                          size->face->glyph->metrics.horiAdvance *
-                          st->scale ),
+                 st->x_origin +
+                   ( ( size->face->glyph->metrics.horiAdvance +
+                       size->face->glyph->lsb_delta           -
+                       size->face->glyph->rsb_delta           ) *
+                     st->scale >> 6 ),
                  0,
                  st->disp_height,
                  st->axis_color );
@@ -586,18 +702,48 @@
       int          nn;
 
 
+      /* render scaled bitmap */
+      if ( st->work & DO_BITMAP )
+      {
+        int             left, top, x_advance, y_advance;
+        grBitmap        bitg;
+        FT_Glyph        glyf;
+
+
+        FT_Get_Glyph( slot, &glyph );
+        error  = FTDemo_Glyph_To_Bitmap( handle, glyph, &bitg, &left, &top,
+                                         &x_advance, &y_advance, &glyf);
+
+        if ( !error )
+        {
+          bitmap_scale( &bitg, scale );
+
+          grBlitGlyphToBitmap( display->bitmap, &bitg,
+                               ox + left * scale, oy - top * scale,
+                               st->axis_color );
+
+          free( bitg.buffer );
+
+          if ( glyf )
+            FT_Done_Glyph( glyf );
+        }
+
+        FT_Done_Glyph( glyph );
+      }
+
       /* scale the outline */
       for ( nn = 0; nn < gimage->n_points; nn++ )
       {
         FT_Vector*  vec = &gimage->points[nn];
 
 
-        vec->x = (FT_F26Dot6)( vec->x * scale );
-        vec->y = (FT_F26Dot6)( vec->y * scale );
+        /* half-pixel shift hints the stroked path */
+        vec->x = vec->x * scale + 32;
+        vec->y = vec->y * scale - 32;
       }
 
       /* stroke then draw it */
-      if ( st->do_outline )
+      if ( st->work & DO_OUTLINE )
       {
         FT_Get_Glyph( slot, &glyph );
         FT_Glyph_Stroke( &glyph, st->stroker, 1 );
@@ -609,12 +755,12 @@
       }
 
       /* draw the points... */
-      if ( st->do_dots )
+      if ( st->work & DO_DOTS )
       {
         for ( nn = 0; nn < gimage->n_points; nn++ )
           circle_draw(
-            (FT_F26Dot6)( st->x_origin * 64 + gimage->points[nn].x ),
-            (FT_F26Dot6)( st->y_origin * 64 - gimage->points[nn].y ),
+            st->x_origin * 64 + gimage->points[nn].x,
+            st->y_origin * 64 - gimage->points[nn].y,
             128,
             handle,
             display,
@@ -623,11 +769,12 @@
       }
 
       /* ... and point numbers */
-      if ( st->do_dotnumbers )
+      if ( st->work & DO_DOTNUMBERS )
       {
         FT_Vector*  points   = gimage->points;
         FT_Short*   contours = gimage->contours;
-        int         cc;
+        char*       tags     = gimage->tags;
+        short       c, n;
         char        number_string[10];
         size_t      number_string_len = sizeof ( number_string );
 
@@ -635,9 +782,9 @@
         FT_Long  octant_y[8] = { 0, 724, 1024, 724, 0, -724, -1024, -724 };
 
 
-        cc = 0;
-        nn = 0;
-        for ( ; cc < gimage->n_contours; cc++ )
+        c = 0;
+        n = 0;
+        for ( ; c < gimage->n_contours; c++ )
         {
           for (;;)
           {
@@ -648,36 +795,36 @@
 
 
             /* find previous and next point in outline */
-            if ( cc == 0 )
+            if ( c == 0 )
             {
-              if ( contours[cc] == 0 )
+              if ( contours[c] == 0 )
               {
                 prev = 0;
                 next = 0;
               }
               else
               {
-                prev = nn > 0 ? nn - 1
-                              : contours[cc];
-                next = nn < contours[cc] ? nn + 1
-                                         : 0;
+                prev = n > 0 ? n - 1
+                             : contours[c];
+                next = n < contours[c] ? n + 1
+                                       : 0;
               }
             }
             else
             {
-              prev = nn > ( contours[cc - 1] + 1 ) ? nn - 1
-                                                   : contours[cc];
-              next = nn < contours[cc] ? nn + 1
-                                       : contours[cc - 1] + 1;
+              prev = n > ( contours[c - 1] + 1 ) ? n - 1
+                                                 : contours[c];
+              next = n < contours[c] ? n + 1
+                                     : contours[c - 1] + 1;
             }
 
             /* get vectors to previous and next point and normalize them; */
             /* we use 16.16 format to improve the computation precision   */
-            in.x = ( points[prev].x - points[nn].x ) * 1024;
-            in.y = ( points[prev].y - points[nn].y ) * 1024;
+            in.x = ( points[prev].x - points[n].x ) * 1024;
+            in.y = ( points[prev].y - points[n].y ) * 1024;
 
-            out.x = ( points[next].x - points[nn].x ) * 1024;
-            out.y = ( points[next].y - points[nn].y ) * 1024;
+            out.x = ( points[next].x - points[n].x ) * 1024;
+            out.y = ( points[next].y - points[n].y ) * 1024;
 
             in_len  = FT_Vector_Length( &in );
             out_len = FT_Vector_Length( &out );
@@ -709,8 +856,8 @@
               {
                 /* use direction based on point index for the offset */
                 /* if we still don't have a good value               */
-                middle.x = octant_x[nn % 8];
-                middle.y = octant_y[nn % 8];
+                middle.x = octant_x[n % 8];
+                middle.y = octant_y[n % 8];
               }
             }
 
@@ -724,7 +871,7 @@
 
             num_digits = snprintf( number_string,
                                    number_string_len,
-                                   "%d", nn );
+                                   "%d", n );
 
             /* we now position the point number in the opposite       */
             /* direction of the `middle' vector, adding some offset   */
@@ -733,17 +880,19 @@
             /* pixels)                                                */
             grWriteCellString( display->bitmap,
                                st->x_origin +
-                                 ( ( points[nn].x - middle.x ) >> 6 ) -
+                                 ( ( points[n].x - middle.x ) >> 6 ) -
                                  ( middle.x > 0 ? ( num_digits - 1 ) * 8 + 2
                                                 : 2 ),
                                st->y_origin -
-                                 ( ( ( points[nn].y - middle.y ) >> 6 ) +
+                                 ( ( ( points[n].y - middle.y ) >> 6 ) +
                                    8 / 2 ),
                                number_string,
-                               st->axis_color );
+                               ( tags[n] & FT_CURVE_TAG_ON )
+                                 ? st->on_color
+                                 : st->off_color );
 
-            nn++;
-            if ( nn > contours[cc] )
+            n++;
+            if ( n > contours[c] )
               break;
           }
         }
@@ -829,23 +978,24 @@
     grWriteln( "p, n        previous/next font          PgUp, PgDn  zoom in/out grid        " );
     grWriteln( "                                        SPC         reset zoom and position " );
 #endif /* FT_DEBUG_AUTOFIT */
-    grWriteln( "Up, Down    adjust size by 0.5pt                                            " );
-    grWriteln( "                                        if not auto-hinting:                " );
-    grWriteln( "Left, Right adjust index by 1             H         cycle through hinting   " );
-    grWriteln( "F7, F8      adjust index by 10                       engines (if available) " );
-    grWriteln( "F9, F10     adjust index by 100         if light auto-hinting:              " );
-    grWriteln( "F11, F12    adjust index by 1000          w         toggle warping          " );
-    grWriteln( "                                                      (if available)        " );
-    grWriteln( "h           toggle hinting                                                  " );
+    grWriteln( "Up, Down    adjust size by 0.5pt        if not auto-hinting:                " );
+    grWriteln( "                                          H         cycle through hinting   " );
+    grWriteln( "Left, Right adjust index by 1                        engines (if available) " );
+    grWriteln( "F7, F8      adjust index by 10          if light auto-hinting:              " );
+    grWriteln( "F9, F10     adjust index by 100           w         toggle warping          " );
+    grWriteln( "F11, F12    adjust index by 1000                      (if available)        " );
+    grWriteln( "                                                                            " );
+    grWriteln( "h           toggle hinting              b           toggle bitmap           " );
     grWriteln( "f           toggle forced auto-         d           toggle dot display      " );
     grWriteln( "             hinting (if hinting)       o           toggle outline display  " );
     grWriteln( "                                        D           toggle dotnumber display" );
-    grWriteln( "a           toggle anti-aliasing        g, v        adjust gamma value      " );
     grWriteln( "                                                                            " );
-    grWriteln( "if Multiple Master or GX font:          q, ESC      quit ftgrid             " );
-    grWriteln( "  F2        cycle through axes                                              " );
-    grWriteln( "  F3, F4    adjust current axis by      F5, F6      cycle through           " );
-    grWriteln( "             1/50th of its range                    anti-aliasing modes     " );
+    grWriteln( "F5, F6      cycle through               if Multiple Master or GX font:      " );
+    grWriteln( "             anti-aliasing modes          F2        cycle through axes      " );
+    grWriteln( "L           cycle through LCD             F3, F4    adjust current axis by  " );
+    grWriteln( "             filters                                 1/50th of its range    " );
+    grWriteln( "g, v        adjust gamma value                                              " );
+    grWriteln( "                                        q, ESC      quit ftgrid             " );
     /*          |----------------------------------|    |----------------------------------| */
     grLn();
     grLn();
@@ -859,7 +1009,7 @@
   static void
   event_cff_hinting_engine_change( int  delta )
   {
-    int  new_cff_hinting_engine;
+    int  new_cff_hinting_engine = 0;
 
 
     if ( delta )
@@ -893,18 +1043,14 @@
   static void
   event_tt_interpreter_version_change( void )
   {
-    FT_UInt  new_interpreter_version;
-
-
-    if ( status.tt_interpreter_version == TT_INTERPRETER_VERSION_35 )
-      new_interpreter_version = TT_INTERPRETER_VERSION_38;
-    else
-      new_interpreter_version = TT_INTERPRETER_VERSION_35;
+    status.tt_interpreter_version_idx += 1;
+    status.tt_interpreter_version_idx %= status.num_tt_interpreter_versions;
 
     error = FT_Property_Set( handle->library,
                              "truetype",
                              "interpreter-version",
-                             &new_interpreter_version );
+                             &status.tt_interpreter_versions[
+                               status.tt_interpreter_version_idx] );
 
     if ( !error )
     {
@@ -912,12 +1058,12 @@
       /* lazy to walk over all loaded fonts to check whether they */
       /* are of type TTF, then unloading them explicitly.         */
       FTC_Manager_Reset( handle->cache_manager );
-      status.tt_interpreter_version = new_interpreter_version;
     }
 
-    sprintf( status.header_buffer, "TrueType engine changed to version %s",
-             status.tt_interpreter_version == TT_INTERPRETER_VERSION_35
-               ? "35" : "38" );
+    sprintf( status.header_buffer,
+             "TrueType engine changed to version %d",
+             status.tt_interpreter_versions[
+               status.tt_interpreter_version_idx]);
 
     status.header = (const char *)status.header_buffer;
   }
@@ -995,8 +1141,16 @@
   static void
   event_grid_zoom( double  zoom )
   {
+    FT_F26Dot6  scale_old = status.scale;
+
+
     status.scale *= zoom;
 
+    /* avoid same zoom value due to truncation */
+    /* to integer in above multiplication      */
+    if ( status.scale == scale_old && zoom > 1.0 )
+      status.scale++;
+
     sprintf( status.header_buffer, "zoom level %.0f%%",
              status.scale * 100.0 / status.scale_0 );
 
@@ -1007,45 +1161,104 @@
   static void
   event_lcd_mode_change( int  delta )
   {
-    if ( handle->antialias )
+    const char*  lcd_mode = NULL;
+
+
+    handle->lcd_mode = ( handle->lcd_mode +
+                         delta            +
+                         N_LCD_MODES      ) % N_LCD_MODES;
+
+    switch ( handle->lcd_mode )
     {
-      const char*  lcd_mode = NULL;
+    case LCD_MODE_MONO:
+      lcd_mode = "monochrome";
+      break;
+    case LCD_MODE_AA:
+      lcd_mode = "normal AA";
+      break;
+    case LCD_MODE_LIGHT:
+      lcd_mode = "light AA";
+      break;
+    case LCD_MODE_LIGHT_SUBPIXEL:
+      lcd_mode = "light AA (subpixel positioning)";
+      break;
+    case LCD_MODE_RGB:
+      lcd_mode = "LCD (horiz. RGB)";
+      break;
+    case LCD_MODE_BGR:
+      lcd_mode = "LCD (horiz. BGR)";
+      break;
+    case LCD_MODE_VRGB:
+      lcd_mode = "LCD (vert. RGB)";
+      break;
+    case LCD_MODE_VBGR:
+      lcd_mode = "LCD (vert. BGR)";
+      break;
+    }
+
+    if ( delta )
+      FTC_Manager_Reset( handle->cache_manager );
+
+    sprintf( status.header_buffer, "rendering mode changed to %s",
+             lcd_mode );
+
+    status.header = (const char *)status.header_buffer;
+
+    FTDemo_Update_Current_Flags( handle );
+  }
 
 
-      handle->lcd_mode = ( handle->lcd_mode +
-                           delta            +
-                           N_LCD_MODES      ) % N_LCD_MODES;
+  static void
+  event_lcd_filter_change( void )
+  {
+    if ( handle->lcd_mode >= LCD_MODE_RGB )
+    {
+      const char*  lcd_filter = NULL;
 
-      switch ( handle->lcd_mode )
+
+      switch( status.lcd_filter )
       {
-      case LCD_MODE_AA:
-        lcd_mode = "normal AA";
+      case FT_LCD_FILTER_DEFAULT:
+        status.lcd_filter = FT_LCD_FILTER_LIGHT;
         break;
-      case LCD_MODE_LIGHT:
-        lcd_mode = "light AA";
+      case FT_LCD_FILTER_LIGHT:
+        status.lcd_filter = FT_LCD_FILTER_LEGACY1;
         break;
-      case LCD_MODE_RGB:
-        lcd_mode = "LCD (horiz. RGB)";
+      case FT_LCD_FILTER_LEGACY1:
+        status.lcd_filter = FT_LCD_FILTER_NONE;
         break;
-      case LCD_MODE_BGR:
-        lcd_mode = "LCD (horiz. BGR)";
-        break;
-      case LCD_MODE_VRGB:
-        lcd_mode = "LCD (vert. RGB)";
-        break;
-      case LCD_MODE_VBGR:
-        lcd_mode = "LCD (vert. BGR)";
+      case FT_LCD_FILTER_NONE:
+      default:
+        status.lcd_filter = FT_LCD_FILTER_DEFAULT;
         break;
       }
-      sprintf( status.header_buffer, "rendering mode changed to %s",
-               lcd_mode );
+
+      switch ( status.lcd_filter )
+      {
+      case FT_LCD_FILTER_DEFAULT:
+        lcd_filter = "default";
+        break;
+      case FT_LCD_FILTER_LIGHT:
+        lcd_filter = "light";
+        break;
+      case FT_LCD_FILTER_LEGACY1:
+        lcd_filter = "legacy";
+        break;
+      case FT_LCD_FILTER_NONE:
+      default:
+        lcd_filter = "none";
+        break;
+      }
+
+      sprintf( status.header_buffer, "LCD filter changed to %s",
+               lcd_filter );
 
       status.header = (const char *)status.header_buffer;
 
-      FTDemo_Update_Current_Flags( handle );
+      FT_Library_SetLcdFilter( handle->library, status.lcd_filter );
     }
     else
-      status.header = "need anti-aliasing to change rendering mode";
+      status.header = "need LCD mode to change filter";
   }
 
 
@@ -1123,6 +1336,7 @@
     FT_Error         err;
     FT_Size          size;
     FT_UInt          n, num_names;
+    FT_Int           instance_index;
     FT_Multi_Master  dummy;
     int              num_indices, is_GX;
 
@@ -1165,6 +1379,9 @@
 
     num_names = FT_Get_Sfnt_Name_Count( size->face );
 
+    /* in `face_index', the instance index starts with value 1 */
+    instance_index = ( size->face->face_index >> 16 ) - 1;
+
     for ( n = 0; n < MAX_MM_AXES; n++ )
     {
       free( status.axis_name[n] );
@@ -1173,7 +1390,21 @@
 
     for ( n = 0; n < status.used_num_axis; n++ )
     {
-      status.design_pos[n] = status.mm->axis[n].def;
+      if ( status.requested_cnt )
+      {
+        status.design_pos[n] = n < status.requested_cnt
+                                 ? status.requested_pos[n]
+                                 : status.mm->axis[n].def;
+        if ( status.design_pos[n] < status.mm->axis[n].minimum )
+          status.design_pos[n] = status.mm->axis[n].minimum;
+        else if ( status.design_pos[n] > status.mm->axis[n].maximum )
+          status.design_pos[n] = status.mm->axis[n].maximum;
+      }
+      else if ( FT_IS_NAMED_INSTANCE( size->face ) )
+        status.design_pos[n] = status.mm->namedstyle[instance_index].
+                                          coords[n];
+      else
+        status.design_pos[n] = status.mm->axis[n].def;
 
       if ( is_GX )
       {
@@ -1269,14 +1500,6 @@
       event_help();
       break;
 
-    case grKEY( 'a' ):
-      handle->antialias = !handle->antialias;
-      status.header     = handle->antialias ? "anti-aliasing is now on"
-                                            : "anti-aliasing is now off";
-
-      FTDemo_Update_Current_Flags( handle );
-      break;
-
     case grKEY( 'f' ):
       handle->autohint = !handle->autohint;
       status.header    = handle->autohint ? "forced auto-hinting is now on"
@@ -1287,8 +1510,10 @@
 
 #ifdef FT_DEBUG_AUTOFIT
     case grKEY( '1' ):
-      if ( handle->hinted                                             &&
-           ( handle->autohint || handle->lcd_mode == LCD_MODE_LIGHT ) )
+      if ( handle->hinted                                  &&
+           ( handle->autohint                            ||
+             handle->lcd_mode == LCD_MODE_LIGHT          ||
+             handle->lcd_mode == LCD_MODE_LIGHT_SUBPIXEL ) )
       {
         status.header = "dumping glyph edges to stdout";
         af_glyph_hints_dump_edges( _af_debug_hints, 1 );
@@ -1296,8 +1521,10 @@
       break;
 
     case grKEY( '2' ):
-      if ( handle->hinted                                             &&
-           ( handle->autohint || handle->lcd_mode == LCD_MODE_LIGHT ) )
+      if ( handle->hinted                                  &&
+           ( handle->autohint                            ||
+             handle->lcd_mode == LCD_MODE_LIGHT          ||
+             handle->lcd_mode == LCD_MODE_LIGHT_SUBPIXEL ) )
       {
         status.header = "dumping glyph segments to stdout";
         af_glyph_hints_dump_segments( _af_debug_hints, 1 );
@@ -1305,8 +1532,10 @@
       break;
 
     case grKEY( '3' ):
-      if ( handle->hinted                                             &&
-           ( handle->autohint || handle->lcd_mode == LCD_MODE_LIGHT ) )
+      if ( handle->hinted                                  &&
+           ( handle->autohint                            ||
+             handle->lcd_mode == LCD_MODE_LIGHT          ||
+             handle->lcd_mode == LCD_MODE_LIGHT_SUBPIXEL ) )
       {
         status.header = "dumping glyph points to stdout";
         af_glyph_hints_dump_points( _af_debug_hints, 1 );
@@ -1314,6 +1543,10 @@
       break;
 #endif /* FT_DEBUG_AUTOFIT */
 
+    case grKEY( 'L' ):
+      event_lcd_filter_change();
+      break;
+
     case grKEY( 'g' ):
       event_gamma_change( 0.1 );
       break;
@@ -1331,19 +1564,24 @@
       status.header  = handle->hinted ? "glyph hinting is now active"
                                       : "glyph hinting is now ignored";
 
+      FTC_Manager_Reset( handle->cache_manager );
       FTDemo_Update_Current_Flags( handle );
       break;
 
     case grKEY( 'd' ):
-      status.do_dots = !status.do_dots;
+      status.work ^= DO_DOTS;
       break;
 
     case grKEY( 'D' ):
-      status.do_dotnumbers = !status.do_dotnumbers;
+      status.work ^= DO_DOTNUMBERS;
       break;
 
     case grKEY( 'o' ):
-      status.do_outline = !status.do_outline;
+      status.work ^= DO_OUTLINE;
+      break;
+
+    case grKEY( 'b' ):
+      status.work ^= DO_BITMAP;
       break;
 
     case grKEY( 'p' ):
@@ -1351,7 +1589,9 @@
       break;
 
     case grKEY( 'H' ):
-      if ( !( handle->autohint || handle->lcd_mode == LCD_MODE_LIGHT ) )
+      if ( !( handle->autohint                            ||
+              handle->lcd_mode == LCD_MODE_LIGHT          ||
+              handle->lcd_mode == LCD_MODE_LIGHT_SUBPIXEL ) )
       {
         FT_Face    face;
         FT_Module  module;
@@ -1385,7 +1625,9 @@
 
 #ifdef FT_DEBUG_AUTOFIT
     case grKEY( 'V' ):
-      if ( handle->autohint || handle->lcd_mode == LCD_MODE_LIGHT )
+      if ( handle->autohint                            ||
+           handle->lcd_mode == LCD_MODE_LIGHT          ||
+           handle->lcd_mode == LCD_MODE_LIGHT_SUBPIXEL )
       {
         status.do_vert_hints = !status.do_vert_hints;
         status.header = status.do_vert_hints ? "vertical hinting enabled"
@@ -1396,7 +1638,9 @@
       break;
 
     case grKEY( 'B' ):
-      if ( handle->autohint || handle->lcd_mode == LCD_MODE_LIGHT )
+      if ( handle->autohint                            ||
+           handle->lcd_mode == LCD_MODE_LIGHT          ||
+           handle->lcd_mode == LCD_MODE_LIGHT_SUBPIXEL )
       {
         status.do_blue_hints = !status.do_blue_hints;
         status.header = status.do_blue_hints ? "blue zone hinting enabled"
@@ -1433,10 +1677,10 @@
 #endif
                         break;
 
-    case grKEY( 'i' ):  event_grid_translate(  0,  1 ); break;
-    case grKEY( 'k' ):  event_grid_translate(  0, -1 ); break;
-    case grKEY( 'l' ):  event_grid_translate(  1,  0 ); break;
+    case grKEY( 'i' ):  event_grid_translate(  0, -1 ); break;
+    case grKEY( 'k' ):  event_grid_translate(  0,  1 ); break;
     case grKEY( 'j' ):  event_grid_translate( -1,  0 ); break;
+    case grKEY( 'l' ):  event_grid_translate(  1,  0 ); break;
 
     case grKeyPageUp:   event_grid_zoom( 1.25     ); break;
     case grKeyPageDown: event_grid_zoom( 1 / 1.25 ); break;
@@ -1589,6 +1833,9 @@
     fprintf( stderr,
       "  -r R      Use resolution R dpi (default: 72dpi).\n"
       "  -f index  Specify first index to display (default: 0).\n"
+      "  -d \"axis1 axis2 ...\"\n"
+      "            Specify the design coordinates for each\n"
+      "            Multiple Master axis at start-up.\n"
       "\n"
       "  -v        Show version."
       "\n" );
@@ -1609,13 +1856,32 @@
 
     while ( 1 )
     {
-      option = getopt( *argc, *argv, "f:h:r:vw:" );
+      option = getopt( *argc, *argv, "d:f:h:r:vw:" );
 
       if ( option == -1 )
         break;
 
       switch ( option )
       {
+      case 'd':
+        {
+          FT_UInt    cnt;
+          FT_Fixed*  pos = status.requested_pos;
+          char*      s   = optarg;
+
+
+          for ( cnt = 0; cnt < MAX_MM_AXES && *s; cnt++ )
+          {
+            pos[cnt] = (FT_Fixed)( strtod( s, &s ) * 65536.0 );
+
+            while ( *s == ' ' )
+              ++s;
+          }
+
+          status.requested_cnt = cnt;
+        }
+        break;
+
       case 'f':
         status.Num = atoi( optarg );
         break;
@@ -1681,8 +1947,12 @@
   main( int    argc,
         char*  argv[] )
   {
-    grEvent  event;
-    int      n;
+    grEvent       event;
+    int           n;
+    unsigned int  dflt_tt_interpreter_version;
+    unsigned int  versions[3] = { TT_INTERPRETER_VERSION_35,
+                                  TT_INTERPRETER_VERSION_38,
+                                  TT_INTERPRETER_VERSION_40 };
 
 
     /* initialize engine */
@@ -1695,13 +1965,32 @@
     FT_Property_Get( handle->library,
                      "cff",
                      "hinting-engine", &status.cff_hinting_engine );
+
+    /* collect all available versions, then set again the default */
     FT_Property_Get( handle->library,
                      "truetype",
-                     "interpreter-version", &status.tt_interpreter_version );
+                     "interpreter-version", &dflt_tt_interpreter_version );
+    for ( n = 0; n < 3; n++ )
+    {
+      error = FT_Property_Set( handle->library,
+                               "truetype",
+                               "interpreter-version", &versions[n] );
+      if ( !error )
+        status.tt_interpreter_versions[
+          status.num_tt_interpreter_versions++] = versions[n];
+      if ( versions[n] == dflt_tt_interpreter_version )
+        status.tt_interpreter_version_idx = n;
+    }
+    FT_Property_Set( handle->library,
+                     "truetype",
+                     "interpreter-version", &dflt_tt_interpreter_version );
+
     FT_Property_Get( handle->library,
                      "autofitter",
                      "warping", &status.warping );
 
+    FT_Library_SetLcdFilter( handle->library, status.lcd_filter );
+
     display = FTDemo_Display_New( gr_pixel_mode_rgb24,
                                   status.width, status.height );
     if ( !display )
@@ -1714,7 +2003,8 @@
 
     for ( ; argc > 0; argc--, argv++ )
     {
-      error = FTDemo_Install_Font( handle, argv[0], 1 );
+      error = FTDemo_Install_Font( handle, argv[0], 1,
+                                   status.requested_cnt ? 1 : 0 );
       if ( error == FT_Err_Invalid_Argument )
         fprintf( stderr, "skipping font `%s' without outlines\n",
                          argv[0] );
@@ -1737,7 +2027,7 @@
 
       grid_status_draw_grid( &status );
 
-      if ( status.do_outline || status.do_dots || status.do_dotnumbers )
+      if ( status.work )
         grid_status_draw_outline( &status, handle, display );
 
       write_header( 0 );
